#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/wait.h>

#define NCELL	10
#define NSEMS	NCELL

union semun {
	int              val;    /* Value for SETVAL */
	struct semid_ds *buf;    /* Buffer for IPC_STAT, IPC_SET */
	unsigned short  *array;  /* Array for GETALL, SETALL */
	struct seminfo  *__buf;  /* Buffer for IPC_INFO
                            (Linux-specific) */
};

void
first_group(int shmid, int semid)
{
	char *buff;
	int i, oflg, res;
	struct sembuf sop;

	buff = shmat(shmid, NULL, 0);
	if (buff == (void*)-1) {
		perror("[-]shmat fail:");
		return;
	}

	oflg = 0;

	for (i = 0; i < NCELL && oflg == 0; i++) {
		sop.sem_num = i;
		sop.sem_op = -1;
		sop.sem_flg = 0;
		res = semop(semid, &sop, 1);

		if (res == -1)
			perror("[-] semop error first group pre:");

		if (buff[i] == 0) {
			printf("[+] found buff[%d]\n", i);
			buff[i] = 1;
			oflg = 1;
		}

		sop.sem_op = 1;
		res = semop(semid, &sop, 1);
		if (res == -1)
			perror("[-] semop error first group post:");
		
	}

	res = shmdt(buff);
	if (res == -1) {
		perror("[-]shmdt fail:");
	}
}

void
second_group(int shmid, int semid)
{
	char *buff;
	int i, oflg, res;
	struct sembuf sop;

	buff = shmat(shmid, NULL, 0);
	if (buff == (void*)-1) {
		perror("[-]shmat fail:");
		return;
	}

	oflg = 0;
	for (i = 0; i < NCELL && oflg == 0; i++) {
		sop.sem_num = i;
		sop.sem_op = -1;
		sop.sem_flg = 0;
		res = semop(semid, &sop, 1);

		if (res == -1) {
			printf("sem_num = %d\n", sop.sem_num);
			perror("[-]semop error second group pre:");
		}

		if (buff[i] != 0 && buff[i] < 10) {
			printf("[+]found buff[%d] < 10; adding\n", i);
			buff[i] += 10;
			printf("  buff[%d] = %d\n", i, buff[i]);
			oflg = 1;
		}
		sop.sem_op = 1;
		res = semop(semid, &sop, 1);
		if (res == -1)
			perror("[-] semop error second group post:");
	}

	res = shmdt(buff);
	if (res == -1) {
		perror("[-]shmdt fail:");
	}
}

void
spawn_childrens(int shmid)
{
	pid_t pid;
	int i, ret, semid;
	unsigned short ptr[NSEMS];
	union semun arg;

	for (i = 0; i < NSEMS; i++)
		ptr[i] = 1;
	arg.array = &ptr;

	semid = semget(IPC_PRIVATE, NSEMS, IPC_CREAT | S_IRWXU);
	if (semid == -1) {
		perror("semget error:");
		return;
	}

	ret = semctl(semid, 0, SETALL, arg);
	if (ret == -1) {
		perror("[-] SEMCTL fail");
		goto finalize;
	}

	for (i = 0; i < NCELL; i++) {
		pid = fork();
		if (pid == 0)
			break;
	}

	if ((i % 2) == 0)
		first_group(shmid, semid);
	else
		second_group(shmid, semid);
	
	//finalize child processes
	if (pid == 0) {
		printf("[+] group %d child exited\n", (i % 2) + 1);
		exit(0);
	}

finalize:
	//need to check for SIGCHILD
	for (i = 0; i < NCELL; i++)
		wait(&ret);

	ret = semctl(semid, 0xDEADBEEF, IPC_RMID);
	if (ret == -1)
		perror("semctl RMID error:");
}


int
main(int argc, char *argv[])
{
	struct shmid_ds sh_descr;
	char *buff;
	int shmid;
	int i, res;
	int pagesize;

	pagesize  = getpagesize();

	shmid = shmget(IPC_PRIVATE, pagesize, IPC_CREAT | S_IRWXU);
	
	if (shmid < 0) {
		printf("[-]shmget fail\n");
		exit(1);
	}

	printf("[+]shmget returns %d\n", shmid);

	buff = shmat(shmid, NULL, 4);
	if (buff == (void*)-1) {
		perror("[-]shmat fail:");
		goto finalize;
	}

	printf("[+] buff is %p\n", buff);
	memset(buff, 0, pagesize);

	spawn_childrens(shmid);

	//Wait for childrens finalize
	while (1) {
		if (shmctl(shmid, IPC_STAT, &sh_descr) == -1) {
			perror("shmctl get stat error:");
			goto finalize;
		}
		if (sh_descr.shm_nattch <= 1)
			break;
		usleep(3000);
	}

	printf("after all:\n");
	for (i = 0; i < NCELL; i++)
		printf("buff[%d]=%d\n", i, buff[i]);

	res = shmdt(buff);
	if (res == -1) {
		perror("[-]shmdt fail:");
	}

finalize:
	
	if (shmctl(shmid, IPC_RMID, NULL) != 0) {
		printf("[-]destroy sh memory segment fail\n");
		exit(1);
	}

	return 0;
}

